#!/usr/bin/env python 
"""
Copyright (C) 2009 Nick Drobchenko, nick@cnc-club.ru
based on gcode.py (C) 2007 hugomatic... 
based on addnodes.py (C) 2005,2007 Aaron Spike, aaron@ekips.org
based on dots.py (C) 2005 Aaron Spike, aaron@ekips.org
based on interp.py (C) 2005 Aaron Spike, aaron@ekips.org
based on bezmisc.py (C) 2005 Aaron Spike, aaron@ekips.org
based on cubicsuperpath.py (C) 2005 Aaron Spike, aaron@ekips.org

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

###
###		Unordinary gears
###
gcodetools_current_version = "1.6.03"

import inkex, simplestyle, simplepath
import cubicsuperpath, simpletransform, bezmisc

import os
import math
import bezmisc
import re
import copy
import sys
import time
import cmath
import numpy
import codecs
import random
import gettext
_ = gettext.gettext

 ### Check if inkex has errormsg (0.46 version doesnot have one.) Could be removed later.
if "errormsg" not in dir(inkex):
	inkex.errormsg = lambda msg: sys.stderr.write((unicode(msg) + "\n").encode("UTF-8"))

def csp_true_bounds(csp) :
	# Finds minx,miny,maxx,maxy of the csp and return their (x,y,i,j,t) 
	minx = [float("inf"), 0, 0, 0]																																																																																																								
	maxx = [float("-inf"), 0, 0, 0]
	miny = [float("inf"), 0, 0, 0]
	maxy = [float("-inf"), 0, 0, 0]
	for i in range(len(csp)):
		for j in range(1,len(csp[i])):
			ax,ay,bx,by,cx,cy,x0,y0 = bezmisc.bezierparameterize((csp[i][j-1][1],csp[i][j-1][2],csp[i][j][0],csp[i][j][1]))
			roots = cubic_solver(0, 3*ax, 2*bx, cx)	 + [0,1]
			for root in roots :
				if type(root) is complex and abs(root.imag)<1e-10:
					root = root.real
				if type(root) is not complex and 0<=root<=1:
					y = ay*(root**3)+by*(root**2)+cy*root+y0  
					x = ax*(root**3)+bx*(root**2)+cx*root+x0  
					maxx = max([x,y,i,j,root],maxx)
					minx = min([x,y,i,j,root],minx)

			roots = cubic_solver(0, 3*ay, 2*by, cy)	 + [0,1]
			for root in roots :
				if type(root) is complex and root.imag==0:
					root = root.real
				if type(root) is not complex and 0<=root<=1:
					y = ay*(root**3)+by*(root**2)+cy*root+y0  
					x = ax*(root**3)+bx*(root**2)+cx*root+x0  
					maxy = max([y,x,i,j,root],maxy)
					miny = min([y,x,i,j,root],miny)
	maxy[0],maxy[1] = maxy[1],maxy[0]
	miny[0],miny[1] = miny[1],miny[0]

	return minx,miny,maxx,maxy
	
def cubic_solver(a,b,c,d):	
	if a!=0:
		#	Monics formula see http://en.wikipedia.org/wiki/Cubic_function#Monic_formula_of_roots
		a,b,c = (b/a, c/a, d/a)
		m = 2*a**3 - 9*a*b + 27*c
		k = a**2 - 3*b
		n = m**2 - 4*k**3
		w1 = -.5 + .5*cmath.sqrt(3)*1j
		w2 = -.5 - .5*cmath.sqrt(3)*1j
		if n>=0 :
			t = m+math.sqrt(n)
			m1 = pow(t/2,1./3) if t>=0 else -pow(-t/2,1./3)
			t = m-math.sqrt(n)
			n1 = pow(t/2,1./3) if t>=0 else -pow(-t/2,1./3)
		else :
			m1 = pow(complex((m+cmath.sqrt(n))/2),1./3)
			n1 = pow(complex((m-cmath.sqrt(n))/2),1./3)
		x1 = -1./3 * (a + m1 + n1)
		x2 = -1./3 * (a + w1*m1 + w2*n1)
		x3 = -1./3 * (a + w2*m1 + w1*n1)
		return [x1,x2,x3]
	elif b!=0:
		det = c**2-4*b*d
		if det>0 :
			return [(-c+math.sqrt(det))/(2*b),(-c-math.sqrt(det))/(2*b)]
		elif d == 0 :
			return [-c/(b*b)] 	
		else :
			return [(-c+cmath.sqrt(det))/(2*b),(-c-cmath.sqrt(det))/(2*b)]
	elif c!=0 :
		return [-d/c]
	else : return []	
	

class Unordinary_gears(inkex.Effect):
	
	def get_transforms(self,g):
		root = self.document.getroot()
		trans = []
		while (g!=root):
			if 'transform' in g.keys():
				t = g.get('transform')
				t = simpletransform.parseTransform(t)
				trans = simpletransform.composeTransform(t,trans) if trans != [] else t
			g=g.getparent()
		return trans 
	

	def apply_transforms(self,g,csp):
		trans = self.get_transforms(g)
		if trans != []:
			simpletransform.applyTransformToPath(trans, csp)
		return csp

	def __init__(self):
		inkex.Effect.__init__(self)
		self.OptionParser.add_option("", "--selected_puley_teeth",		action="store", type="int", 		dest="selected_puley_teeth", default="10",		help="Selected pulley teeth")
		self.OptionParser.add_option("", "--generated_puley_teeth",		action="store", type="int", 		dest="generated_puley_teeth", default="20",		help="Generated pulley teeth")
		self.OptionParser.add_option("", "--number_of_copies",			action="store", type="int", 		dest="number_of_copies", default="50",			help="Number of copies")
		self.OptionParser.add_option("", "--distance",					action="store", type="float", 		dest="distance", default="100",					help="Center to center distance")
		self.OptionParser.add_option("", "--units",						action="store", type="float", 		dest="units", default="90",					help="Units")
		self.OptionParser.add_option("", "--active_tab",				action="store", type="string", 		dest="active_tab", default="",					help="")

	def error(self, s, type_= "Warning"):
		warnings = """
						Warning						
						"""
		errors = """
						Error 	
					"""
		if type_.lower() in re.split("[\s\n,\.]+", errors.lower()) :
			inkex.errormsg(s+"\n")		
			sys.exit()
		elif type_.lower() in re.split("[\s\n,\.]+", warnings.lower()) :
			inkex.errormsg(s+"\n")		
		else :
			inkex.errormsg(s)		
			sys.exit()

	def draw_pointer(self, x,color = "#f00", figure = "cross", comment = "", width = .1) :
		if figure ==  "line" :
			s = ""
			for i in range(1,len(x)/2) :
				s+= " %s, %s " %(x[i*2],x[i*2+1])
			inkex.etree.SubElement( self.document.getroot(), inkex.addNS('path','svg'), {"d": "M %s,%s L %s"%(x[0],x[1],s), "style":"fill:none;stroke:%s;stroke-width:%f;"%(color,width),"comment":str(comment)} )
		else :
			inkex.etree.SubElement( self.document.getroot(), inkex.addNS('path','svg'), {"d": "m %s,%s l 10,10 -20,-20 10,10 -10,10, 20,-20"%(x[0],x[1]), "style":"fill:none;stroke:%s;stroke-width:%f;"%(color,width),"comment":str(comment)} )
	
			
										
################################################################################
###
###		Effect
###
###		Main function
###
################################################################################
	def effect(self) :
		self.distance = self.options.distance * self.options.units
		self.first_teeth = self.options.selected_puley_teeth
		self.second_teeth = self.options.generated_puley_teeth
		self.num = self.options.number_of_copies

		if len(self.selected)!=1 : self.error('Please select one and only one path!','error')
		path = self.selected.values()[0]
		if "d" not in path.keys() :  self.error('Please select a "Path"!','error')
		csp = cubicsuperpath.parsePath(path.get("d"))
		csp = self.apply_transforms(path, csp)
		if "transform" in path.keys() :	del path.attrib["transform"]

		path.set('d', cubicsuperpath.formatPath(csp))

				# Will have to find center of bbox
		minx,miny,maxx,maxy = csp_true_bounds(csp)

		c1x = (maxx[0] + minx[0])/2
		c1y = (maxy[1] + miny[1])/2
		if inkex.addNS('transform-center-x','inkscape') in path.keys() : c1x += float(path.get(inkex.addNS('transform-center-x','inkscape')))
		if inkex.addNS('transform-center-y','inkscape') in path.keys() : c1y -= float(path.get(inkex.addNS('transform-center-y','inkscape')))
		
		c2x,c2y = c1x-self.distance, c1y

		self.draw_pointer([c1x,c1y], color = "#000080", width = 1 )
		self.draw_pointer([c2x,c2y], color = "#000080", width = 1 )
		
		group = inkex.etree.SubElement( path.getparent(),  inkex.addNS('g','svg') )
		attrib = path.attrib
		for i in range(self.num):
			alpha = math.pi*2*(i)/self.num
			alpha1 = alpha*self.second_teeth# + alpha*self.first_teeth 
			alpha2 = alpha*self.first_teeth 

			# move c1 to 0 
			transform = [ [1,0,-c1x], [0,1,-c1y] ]
			# Rotate to alpha1 to 0 
			transform = simpletransform.composeTransform([ [math.cos(alpha1), -math.sin(alpha1), 0], [math.sin(alpha1), math.cos(alpha1), 0] ], transform )
			# Move c2 to 0 			
			transform = simpletransform.composeTransform( [ [1,0,-c2x+c1x], [0,1,-c2y+c1y] ], transform)
			# Rotate to alpha2 to 0 
			transform = simpletransform.composeTransform( [ [math.cos(alpha2), -math.sin(alpha2), 0], [math.sin(alpha2), math.cos(alpha2), 0] ] , transform)
			# move c2 back to c2
			transform = simpletransform.composeTransform([ [1,0,c2x], [0,1,c2y] ], transform)
			
			# add path's clone 
			attrib["transform"] = simpletransform.formatTransform(transform)
			inkex.etree.SubElement( group, inkex.addNS('path','svg'), attrib )							

#						
e = Unordinary_gears()
e.affect()					

